@@ -6,6 +6,7 @@
 #include "ostream.h"
 #include "restrict-access.h"
 #include "master-service.h"
+#include "settings.h"
 #include "mailbox-list-private.h"
 #include "mbox-storage.h"
 #include "mbox-lock.h"
@@ -14,24 +15,18 @@
 #include "istream-raw-mbox.h"
 #include "mail-copy.h"
 #include "index-mail.h"
+#include "doc.h"
 
 #include <sys/stat.h>
 
 /* How often to touch the dotlock file when using KEEP_LOCKED flag */
 #define MBOX_LOCK_TOUCH_MSECS (10*1000)
 
-/* Assume that if atime < mtime, there are new mails. If it's good enough for
-   UW-IMAP, it's good enough for us. */
-#define STAT_GET_MARKED(st) \
-	((st).st_size == 0 ? MAILBOX_UNMARKED : \
-	 (st).st_atime < (st).st_mtime ? MAILBOX_MARKED : MAILBOX_UNMARKED)
-
 #define MBOX_LIST_CONTEXT(obj) \
 	MODULE_CONTEXT_REQUIRE(obj, mbox_mailbox_list_module)
 
 struct mbox_mailbox_list {
 	union mailbox_list_module_context module_ctx;
-	const struct mbox_settings *set;
 };
 
 /* NOTE: must be sorted for istream-header-filter. Note that it's not such
@@ -132,7 +127,7 @@ mbox_list_get_path(struct mailbox_list *list, const char *name,
 	case MAILBOX_LIST_PATH_TYPE_INDEX_CACHE:
 	case MAILBOX_LIST_PATH_TYPE_LIST_INDEX:
 		if (name == NULL && type == MAILBOX_LIST_PATH_TYPE_CONTROL &&
-		    list->set.control_dir != NULL) {
+		    list->mail_set->mail_control_path[0] != '\0') {
 			/* kind of a kludge for backwards compatibility:
 			   the subscriptions file is in the root control_dir
 			   without .imap/ suffix */
@@ -158,6 +153,8 @@ mbox_list_get_path(struct mailbox_list *list, const char *name,
 	case MAILBOX_LIST_PATH_TYPE_INDEX_PRIVATE:
 		*path_r = path;
 		break;
+	case MAILBOX_LIST_PATH_TYPE_COUNT:
+		i_unreached();
 	}
 	return 1;
 }
@@ -188,192 +185,167 @@ mbox_storage_create(struct mail_storage *_storage, struct mail_namespace *ns,
 		return -1;
 	}
 
-	storage->set = mail_namespace_get_driver_settings(ns, _storage);
+	if (settings_get(_storage->event, &mbox_setting_parser_info, 0,
+			 &storage->set, error_r) < 0)
+		return -1;
 
 	if (mailbox_list_get_root_path(ns->list, MAILBOX_LIST_PATH_TYPE_INDEX, &dir)) {
 		_storage->temp_path_prefix = p_strconcat(_storage->pool, dir,
 			"/", mailbox_list_get_temp_prefix(ns->list), NULL);
 	}
-	if (stat(ns->list->set.root_dir, &st) == 0 && !S_ISDIR(st.st_mode)) {
+	if (stat(ns->list->mail_set->mail_path, &st) == 0 &&
+	    !S_ISDIR(st.st_mode)) {
 		*error_r = t_strdup_printf(
 			"mbox root directory can't be a file: %s "
-			"(http://wiki2.dovecot.org/MailLocation/Mbox)",
-			ns->list->set.root_dir);
+			"(" DOC_LINK("core/config/mailbox/mail_location.html#mbox") ")",
+			ns->list->mail_set->mail_path);
 		return -1;
 	}
 	return 0;
 }
 
-static void mbox_storage_get_list_settings(const struct mail_namespace *ns,
-					   struct mailbox_list_settings *set)
+static void mbox_storage_destroy(struct mail_storage *_storage)
 {
-	if (set->layout == NULL)
-		set->layout = MAILBOX_LIST_NAME_FS;
-	if (set->subscription_fname == NULL)
-		set->subscription_fname = MBOX_SUBSCRIPTION_FILE_NAME;
+	struct mbox_storage *storage = MBOX_STORAGE(_storage);
 
-	if (set->inbox_path == NULL &&
-	    strcasecmp(set->layout, MAILBOX_LIST_NAME_FS) == 0) {
-		set->inbox_path = t_strconcat(set->root_dir, "/inbox", NULL);
-		e_debug(ns->user->event, "mbox: INBOX defaulted to %s", set->inbox_path);
-	}
+	settings_free(storage->set);
+	index_storage_destroy(_storage);
 }
 
-static bool mbox_is_file(const char *path, const char *name, bool debug)
+static bool mbox_is_file(const char *path, const char *name, struct event *event)
 {
 	struct stat st;
 
 	if (stat(path, &st) < 0) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: stat(%s) failed: %m",
-				name, path);
-		}
+		e_debug(event, "mbox autodetect: %s: stat(%s) failed: %m",
+			name, path);
 		return FALSE;
 	}
 	if (S_ISDIR(st.st_mode)) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: is a directory (%s)",
-			       name, path);
-		}
+		e_debug(event, "mbox autodetect: %s: is a directory (%s)",
+			name, path);
 		return FALSE;
 	}
-	if (access(path, R_OK|W_OK) < 0) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: no R/W access (%s)",
-			       name, path);
-		}
+	if (i_faccessat2(AT_FDCWD, path, R_OK | W_OK, AT_EACCESS) < 0) {
+		e_debug(event, "mbox autodetect: %s: no R/W access (%s)",
+			name, path);
 		return FALSE;
 	}
 
-	if (debug)
-		i_debug("mbox autodetect: %s: yes (%s)", name, path);
+	e_debug(event, "mbox autodetect: %s: yes (%s)", name, path);
 	return TRUE;
 }
 
-static bool mbox_is_dir(const char *path, const char *name, bool debug)
+static bool mbox_is_dir(const char *path, const char *name, struct event *event)
 {
 	struct stat st;
 
 	if (stat(path, &st) < 0) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: stat(%s) failed: %m",
-			       name, path);
-		}
+		e_debug(event, "mbox autodetect: %s: stat(%s) failed: %m",
+			name, path);
 		return FALSE;
 	}
 	if (!S_ISDIR(st.st_mode)) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: is not a directory (%s)",
-			       name, path);
-		}
+		e_debug(event, "mbox autodetect: %s: is not a directory (%s)",
+			name, path);
 		return FALSE;
 	}
-	if (access(path, R_OK|W_OK|X_OK) < 0) {
-		if (debug) {
-			i_debug("mbox autodetect: %s: no R/W/X access (%s)",
-			       name, path);
-		}
+	if (i_faccessat2(AT_FDCWD, path, R_OK | W_OK | X_OK, AT_EACCESS) < 0) {
+		e_debug(event, "mbox autodetect: %s: no R/W/X access (%s)",
+			name, path);
 		return FALSE;
 	}
 
-	if (debug)
-		i_debug("mbox autodetect: %s: yes (%s)", name, path);
+	e_debug(event, "mbox autodetect: %s: yes (%s)", name, path);
 	return TRUE;
 }
 
-static bool mbox_storage_is_root_dir(const char *dir, bool debug)
+static bool mbox_storage_is_root_dir(const char *dir, struct event *event)
 {
 	if (mbox_is_dir(t_strconcat(dir, "/"MBOX_INDEX_DIR_NAME, NULL),
-			"has "MBOX_INDEX_DIR_NAME"/", debug))
+			"has "MBOX_INDEX_DIR_NAME"/", event))
 		return TRUE;
-	if (mbox_is_file(t_strconcat(dir, "/inbox", NULL), "has inbox", debug))
+	if (mbox_is_file(t_strconcat(dir, "/inbox", NULL), "has inbox", event))
 		return TRUE;
-	if (mbox_is_file(t_strconcat(dir, "/mbox", NULL), "has mbox", debug))
+	if (mbox_is_file(t_strconcat(dir, "/mbox", NULL), "has mbox", event))
 		return TRUE;
 	return FALSE;
 }
 
 static const char *mbox_storage_find_root_dir(const struct mail_namespace *ns)
 {
-	bool debug = ns->mail_set->mail_debug;
+	struct event *event = ns->user->event;
 	const char *home, *path;
 
 	if (ns->owner == NULL ||
 	    mail_user_get_home(ns->owner, &home) <= 0) {
-		if (debug)
-			i_debug("maildir: Home directory not set");
+		e_debug(event, "mbox autodetect: Home directory not set");
 		home = "";
 	}
 
 	path = t_strconcat(home, "/mail", NULL);
-	if (mbox_storage_is_root_dir(path, debug))
+	if (mbox_storage_is_root_dir(path, event))
 		return path;
 
 	path = t_strconcat(home, "/Mail", NULL);
-	if (mbox_storage_is_root_dir(path, debug))
+	if (mbox_storage_is_root_dir(path, event))
 		return path;
 	return NULL;
 }
 
 static const char *
-mbox_storage_find_inbox_file(const char *user, bool debug)
+mbox_storage_find_inbox_file(struct mail_user *user, struct event *event)
 {
 	const char *path;
 
-	path = t_strconcat("/var/mail/", user, NULL);
-	if (access(path, R_OK|W_OK) == 0) {
-		if (debug)
-			i_debug("mbox: INBOX exists (%s)", path);
+	path = t_strconcat("/var/mail/", user->username, NULL);
+	if (i_faccessat2(AT_FDCWD, path, R_OK | W_OK, AT_EACCESS) == 0) {
+		e_debug(event, "mbox autodetect: INBOX exists (%s)", path);
 		return path;
 	}
-	if (debug)
-		i_debug("mbox: INBOX: access(%s, rw) failed: %m", path);
+	e_debug(event, "mbox autodetect: INBOX: access(%s, rw) failed: %m", path);
 
 	path = t_strconcat("/var/spool/mail/", user, NULL);
-	if (access(path, R_OK|W_OK) == 0) {
-		if (debug)
-			i_debug("mbox: INBOX exists (%s)", path);
+	if (i_faccessat2(AT_FDCWD, path, R_OK | W_OK, AT_EACCESS) == 0) {
+		e_debug(event, "INBOX exists (%s)", path);
 		return path;
 	}
-	if (debug)
-		i_debug("mbox: INBOX: access(%s, rw) failed: %m", path);
+	e_debug(event, "mbox autodetect: INBOX: access(%s, rw) failed: %m", path);
 	return NULL;
 }
 
-static bool mbox_storage_autodetect(const struct mail_namespace *ns,
-				    struct mailbox_list_settings *set)
+static bool
+mbox_storage_autodetect(const struct mail_namespace *ns,
+			const struct mail_storage_settings *mail_set,
+			const char **root_path_r, const char **inbox_path_r)
 {
-	bool debug = ns->mail_set->mail_debug;
+	struct event *event = ns->user->event;
 	const char *root_dir, *inbox_path;
 
-	root_dir = set->root_dir;
-	inbox_path = set->inbox_path;
+	root_dir = mail_set->mail_path;
+	inbox_path = mail_set->mail_inbox_path;
 
-	if (root_dir != NULL) {
+	if (root_dir[0] != '\0') {
 		if (inbox_path == NULL &&
-		    mbox_is_file(root_dir, "INBOX file", debug)) {
+		    mbox_is_file(root_dir, "INBOX file", event)) {
 			/* using location=<INBOX> */
 			inbox_path = root_dir;
-			root_dir = NULL;
-		} else if (!mbox_storage_is_root_dir(root_dir, debug))
+			root_dir = "";
+		} else if (!mbox_storage_is_root_dir(root_dir, event))
 			return FALSE;
 	}
-	if (root_dir == NULL) {
+	if (root_dir[0] == '\0') {
 		root_dir = mbox_storage_find_root_dir(ns);
 		if (root_dir == NULL) {
-			if (debug)
-				i_debug("mbox: couldn't find root dir");
+			e_debug(event, "mbox autodetect: couldn't find root dir");
 			return FALSE;
 		}
 	}
 	if (inbox_path == NULL) {
-		inbox_path = mbox_storage_find_inbox_file(ns->user->username,
-							  debug);
+		inbox_path = mbox_storage_find_inbox_file(ns->user, event);
 	}
-	set->root_dir = root_dir;
-	set->inbox_path = inbox_path;
-
-	mbox_storage_get_list_settings(ns, set);
+	*root_path_r = root_dir;
+	*inbox_path_r = inbox_path;
 	return TRUE;
 }
 
@@ -466,7 +438,7 @@ static int mbox_mailbox_open_existing(struct mbox_mailbox *mbox)
 		   /var/mail and we want to allow privileged dotlocking */
 		rootdir = mailbox_list_get_root_forced(box->list,
 						       MAILBOX_LIST_PATH_TYPE_DIR);
-		if (!str_begins(box_path, rootdir))
+		if (!str_begins_with(box_path, rootdir))
 			mbox->mbox_privileged_locking = TRUE;
 	}
 	if ((box->flags & MAILBOX_FLAG_KEEP_LOCKED) != 0) {
@@ -564,7 +536,7 @@ static int create_inbox(struct mailbox *box)
 	inbox_path = mailbox_get_path(box);
 
 	fd = open(inbox_path, O_RDWR | O_CREAT | O_EXCL, 0660);
-	if (fd == -1 && errno == EACCES) {
+	if (fd == -1 && ENOACCESS(errno)) {
 		/* try again with increased privileges */
 		(void)restrict_access_use_priv_gid();
 		fd = open(inbox_path, O_RDWR | O_CREAT | O_EXCL, 0660);
@@ -573,7 +545,7 @@ static int create_inbox(struct mailbox *box)
 	if (fd != -1) {
 		i_close_fd(&fd);
 		return 0;
-	} else if (errno == EACCES) {
+	} else if (ENOACCESS(errno)) {
 		mailbox_set_critical(box, "%s",
 			mail_error_create_eacces_msg("open", inbox_path));
 		return -1;
@@ -734,16 +706,15 @@ mbox_is_internal_name(struct mailbox_list *list ATTR_UNUSED,
 	return strcmp(name, MBOX_INDEX_DIR_NAME) == 0;
 }
 
-static void mbox_storage_add_list(struct mail_storage *storage,
+static void mbox_storage_add_list(struct mail_storage *storage ATTR_UNUSED,
 				  struct mailbox_list *list)
 {
 	struct mbox_mailbox_list *mlist;
 
 	mlist = p_new(list->pool, struct mbox_mailbox_list, 1);
 	mlist->module_ctx.super = list->v;
-	mlist->set = mail_namespace_get_driver_settings(list->ns, storage);
 
-	if (*list->set.maildir_name == '\0') {
+	if (list->mail_set->mailbox_directory_name[0] == '\0') {
 		/* have to use .imap/ directories */
 		list->v.get_path = mbox_list_get_path;
 	}
@@ -832,8 +803,8 @@ bool mbox_is_backend_readonly(struct mbox_mailbox *mbox)
 {
 	if (!mbox->backend_readonly_set) {
 		mbox->backend_readonly_set = TRUE;
-		if (access(mailbox_get_path(&mbox->box), R_OK|W_OK) < 0 &&
-		    errno == EACCES)
+		if (i_faccessat2(AT_FDCWD, mailbox_get_path(&mbox->box), R_OK | W_OK, AT_EACCESS) < 0 &&
+		    ENOACCESS(errno))
 			mbox->backend_readonly = TRUE;
 	}
 	return mbox->backend_readonly;
@@ -845,14 +816,13 @@ struct mail_storage mbox_storage = {
 		MAIL_STORAGE_CLASS_FLAG_OPEN_STREAMS |
 		MAIL_STORAGE_CLASS_FLAG_HAVE_MAIL_GUIDS,
 	.event_category = &event_category_mbox,
+	.set_info = &mbox_setting_parser_info,
 
 	.v = {
-                mbox_get_setting_parser_info,
 		mbox_storage_alloc,
 		mbox_storage_create,
-		index_storage_destroy,
+		mbox_storage_destroy,
 		mbox_storage_add_list,
-		mbox_storage_get_list_settings,
 		mbox_storage_autodetect,
 		mbox_mailbox_alloc,
 		NULL,
